@@ -4,11 +4,11 @@
 package core
 
 import (
-	"fmt"
 	"github.com/daveshanley/vacuum/model"
 	vacuumUtils "github.com/daveshanley/vacuum/utils"
+	"github.com/pb33f/doctor/model/high/v3"
 	"github.com/pb33f/libopenapi/utils"
-	"gopkg.in/yaml.v3"
+	"go.yaml.in/yaml/v4"
 	"sort"
 )
 
@@ -23,6 +23,11 @@ func (t Truthy) GetSchema() model.RuleFunctionSchema {
 	}
 }
 
+// GetCategory returns the category of the Truthy rule.
+func (t Truthy) GetCategory() string {
+	return model.FunctionCategoryCore
+}
+
 // RunRule will execute the Truthy rule, based on supplied context and a supplied []*yaml.Node slice.
 func (t *Truthy) RunRule(nodes []*yaml.Node, context model.RuleFunctionContext) []model.RuleFunctionResult {
 
@@ -56,12 +61,13 @@ func (t *Truthy) RunRule(nodes []*yaml.Node, context model.RuleFunctionContext)
 			fieldNodeValue.Value == "0" || fieldNodeValue.Value == "" {
 
 			if isArray {
-				pathValue = fmt.Sprintf("%s[%d]", pathValue, x)
+				pathValue = model.GetStringTemplates().BuildArrayPath(pathValue, x)
 			}
 
 			if !utils.IsNodeMap(fieldNode) && !utils.IsNodeArray(fieldNodeValue) && !utils.IsNodeMap(fieldNodeValue) {
 				if context.Index != nil {
 					origin := context.Index.FindNodeOrigin(node)
+
 					if origin != nil && origin.Line > 1 {
 						nm := context.Index.GetNodeMap()
 						var keys []int
@@ -79,14 +85,44 @@ func (t *Truthy) RunRule(nodes []*yaml.Node, context model.RuleFunctionContext)
 						}
 					}
 				}
-				results = append(results, model.RuleFunctionResult{
+
+				var locatedObjects []v3.Foundational
+				var allPaths []string
+				var err error
+				locatedPath := pathValue
+				if context.DrDocument != nil {
+					if fieldNode == nil {
+						locatedObjects, err = context.DrDocument.LocateModel(node)
+					} else {
+						locatedObjects, err = context.DrDocument.LocateModelsByKeyAndValue(fieldNode, fieldNodeValue)
+					}
+					if err == nil && locatedObjects != nil {
+						for x, obj := range locatedObjects {
+							p := model.GetStringTemplates().BuildJSONPath(obj.GenerateJSONPath(), context.RuleAction.Field)
+							if x == 0 {
+								locatedPath = p
+							}
+							allPaths = append(allPaths, p)
+						}
+					}
+				}
+				result := model.RuleFunctionResult{
 					Message: vacuumUtils.SuppliedOrDefault(message,
-						fmt.Sprintf("%s: `%s` must be set", ruleMessage, context.RuleAction.Field)),
+						model.GetStringTemplates().BuildFieldValidationMessage(ruleMessage, context.RuleAction.Field, "set")),
 					StartNode: node,
 					EndNode:   vacuumUtils.BuildEndNode(node),
-					Path:      pathValue,
+					Path:      locatedPath,
 					Rule:      context.Rule,
-				})
+				}
+				if len(allPaths) > 1 {
+					result.Paths = allPaths
+				}
+				results = append(results, result)
+				if len(locatedObjects) > 0 {
+					if arr, ok := locatedObjects[0].(v3.AcceptsRuleResults); ok {
+						arr.AddRuleFunctionResult(v3.ConvertRuleResult(&result))
+					}
+				}
 			}
 		}
 	}
