@@ -1,7 +1,10 @@
 package core
 
 import (
+	"fmt"
 	"github.com/daveshanley/vacuum/model"
+	drModel "github.com/pb33f/doctor/model"
+	"github.com/pb33f/libopenapi"
 	"github.com/pb33f/libopenapi/utils"
 	"github.com/stretchr/testify/assert"
 	"testing"
@@ -20,11 +23,22 @@ func TestUndefined_RunRule(t *testing.T) {
 
 func TestUndefined_RunRule_Success(t *testing.T) {
 
-	sampleYaml := `pizza:
+	sampleYaml :=
+		`openapi: 3.0.0
+pizza:
   cake: "fridge"`
 
 	path := "$.pizza"
 
+	document, err := libopenapi.NewDocument([]byte(sampleYaml))
+	if err != nil {
+		panic(fmt.Sprintf("cannot create new document: %e", err))
+	}
+
+	m, _ := document.BuildV3Model()
+
+	drDocument := drModel.NewDrDocument(m)
+
 	nodes, _ := utils.FindNodes([]byte(sampleYaml), path)
 	assert.Len(t, nodes, 1)
 
@@ -32,6 +46,8 @@ func TestUndefined_RunRule_Success(t *testing.T) {
 	ctx := buildCoreTestContext(model.CastToRuleAction(rule.Then), nil)
 	ctx.Given = path
 	ctx.Rule = &rule
+	ctx.Document = document
+	ctx.DrDocument = drDocument
 
 	def := Undefined{}
 	res := def.RunRule(nodes, ctx)
@@ -41,11 +57,22 @@ func TestUndefined_RunRule_Success(t *testing.T) {
 
 func TestUndefined_RunRule_Fail(t *testing.T) {
 
-	sampleYaml := `pizza:
+	sampleYaml :=
+		`openapi: 3.0.0
+pizza:
   noCake: "noFun"`
 
 	path := "$.pizza"
 
+	document, err := libopenapi.NewDocument([]byte(sampleYaml))
+	if err != nil {
+		panic(fmt.Sprintf("cannot create new document: %e", err))
+	}
+
+	m, _ := document.BuildV3Model()
+
+	drDocument := drModel.NewDrDocument(m)
+
 	nodes, _ := utils.FindNodes([]byte(sampleYaml), path)
 	assert.Len(t, nodes, 1)
 
@@ -53,6 +80,8 @@ func TestUndefined_RunRule_Fail(t *testing.T) {
 	ctx := buildCoreTestContext(model.CastToRuleAction(rule.Then), nil)
 	ctx.Given = path
 	ctx.Rule = &rule
+	ctx.Document = document
+	ctx.DrDocument = drDocument
 
 	def := Undefined{}
 	res := def.RunRule(nodes, ctx)
