@@ -9,26 +9,22 @@ import (
 	vacuumUtils "github.com/daveshanley/vacuum/utils"
 	"github.com/pb33f/libopenapi/index"
 	"github.com/pb33f/libopenapi/utils"
-	"gopkg.in/yaml.v3"
+	"go.yaml.in/yaml/v4"
 )
 
 // OperationSecurityDefined is a rule that checks operation security against defined global schemes.
 type OperationSecurityDefined struct {
 }
 
+// GetCategory returns the category of the OperationSecurityDefined rule.
+func (osd OperationSecurityDefined) GetCategory() string {
+	return model.FunctionCategoryOpenAPI
+}
+
 // GetSchema returns a model.RuleFunctionSchema defining the schema of the OperationSecurityDefined rule.
 func (osd OperationSecurityDefined) GetSchema() model.RuleFunctionSchema {
 	return model.RuleFunctionSchema{
-		Name: "operation_security_defined",
-		Properties: []model.RuleFunctionProperty{
-			{
-				Name:        "schemesPath",
-				Description: "operation_security_defined requires a schemesPath in which to look up security definitions",
-			},
-		},
-		MinProperties: 1,
-		MaxProperties: 1,
-		ErrorMessage:  "operation_security_defined requires a 'schemesPath'",
+		Name: "oasOpSecurityDefined",
 	}
 }
 
