@@ -7,7 +7,7 @@ import (
 	"fmt"
 	"github.com/daveshanley/vacuum/model"
 	vacuumUtils "github.com/daveshanley/vacuum/utils"
-	"gopkg.in/yaml.v3"
+	"go.yaml.in/yaml/v4"
 )
 
 // PolymorphicAnyOf checks that there is no polymorphism used, in particular 'anyOf'
@@ -17,10 +17,15 @@ type PolymorphicAnyOf struct {
 // GetSchema returns a model.RuleFunctionSchema defining the schema of the PolymorphicAnyOf rule.
 func (pm PolymorphicAnyOf) GetSchema() model.RuleFunctionSchema {
 	return model.RuleFunctionSchema{
-		Name: "polymorphic_anyOf",
+		Name: "oasPolymorphicAnyOf",
 	}
 }
 
+// GetCategory returns the category of the PolymorphicAnyOf rule.
+func (pm PolymorphicAnyOf) GetCategory() string {
+	return model.FunctionCategoryOpenAPI
+}
+
 // RunRule will execute the PolymorphicAnyOf rule, based on supplied context and a supplied []*yaml.Node slice.
 func (pm PolymorphicAnyOf) RunRule(nodes []*yaml.Node, context model.RuleFunctionContext) []model.RuleFunctionResult {
 
