@@ -9,7 +9,7 @@ import (
 	vacuumUtils "github.com/daveshanley/vacuum/utils"
 	"github.com/pb33f/libopenapi/index"
 	"github.com/pb33f/libopenapi/utils"
-	"gopkg.in/yaml.v3"
+	"go.yaml.in/yaml/v4"
 	"strings"
 )
 
@@ -20,10 +20,15 @@ type UnusedComponent struct {
 // GetSchema returns a model.RuleFunctionSchema defining the schema of the UnusedComponent rule.
 func (uc UnusedComponent) GetSchema() model.RuleFunctionSchema {
 	return model.RuleFunctionSchema{
-		Name: "unused_component",
+		Name: "oasUnusedComponent",
 	}
 }
 
+// GetCategory returns the category of the UnusedComponent rule.
+func (uc UnusedComponent) GetCategory() string {
+	return model.FunctionCategoryOpenAPI
+}
+
 // RunRule will execute the UnusedComponent rule, based on supplied context and a supplied []*yaml.Node slice.
 func (uc UnusedComponent) RunRule(nodes []*yaml.Node, context model.RuleFunctionContext) []model.RuleFunctionResult {
 
@@ -33,8 +38,12 @@ func (uc UnusedComponent) RunRule(nodes []*yaml.Node, context model.RuleFunction
 
 	var results []model.RuleFunctionResult
 
-	// extract all references, and every single component
-	allRefs := context.Index.GetAllReferences()
+	if context.Document == nil || context.Document.GetRolodex() == nil {
+		return nil
+	}
+
+	// extract all references, and every single component, recursively
+	allRefs := context.Document.GetRolodex().GetAllReferences()
 	schemas := context.Index.GetAllComponentSchemas()
 	responses := context.Index.GetAllResponses()
 	parameters := context.Index.GetAllParameters()
@@ -44,7 +53,8 @@ func (uc UnusedComponent) RunRule(nodes []*yaml.Node, context model.RuleFunction
 	securitySchemes := context.Index.GetAllSecuritySchemes()
 	links := context.Index.GetAllLinks()
 	callbacks := context.Index.GetAllCallbacks()
-	mappedRefs := context.Index.GetMappedReferences()
+	pathItems := context.Index.GetAllComponentPathItems()
+	mappedRefs := context.Document.GetRolodex().GetAllMappedReferences()
 
 	// extract securityRequirements from swagger. These are not mapped as they are not $refs
 	// so, we need to map them as if they were.
@@ -59,6 +69,9 @@ func (uc UnusedComponent) RunRule(nodes []*yaml.Node, context model.RuleFunction
 	checkOpenAPISecurity := func(key string) bool {
 		if strings.Contains(key, "securitySchemes") {
 			segs := strings.Split(key, "/")
+			if len(segs) == 0 {
+				return false
+			}
 			def := segs[len(segs)-1]
 			for r := range context.Index.GetSecurityRequirementReferences() {
 				if r == def {
@@ -97,6 +110,7 @@ func (uc UnusedComponent) RunRule(nodes []*yaml.Node, context model.RuleFunction
 		requestBodies,
 		headers,
 		securitySchemes,
+		pathItems,
 		links,
 		callbacks,
 	}
