@@ -7,29 +7,30 @@ import (
 	"context"
 	"errors"
 	"fmt"
-	vacuumUtils "github.com/daveshanley/vacuum/utils"
-	v3 "github.com/pb33f/libopenapi/datamodel/high/v3"
-	"github.com/sourcegraph/conc"
 	"io"
+	"io/fs"
 	"log/slog"
 	"net/url"
-	"os"
 	"path/filepath"
+	"strconv"
 	"strings"
 	"sync"
 	"time"
 
+	vacuumUtils "github.com/daveshanley/vacuum/utils"
+	v3 "github.com/pb33f/libopenapi/datamodel/high/v3"
+	"github.com/sourcegraph/conc"
+
 	"github.com/daveshanley/vacuum/functions"
 	"github.com/daveshanley/vacuum/model"
 	"github.com/daveshanley/vacuum/rulesets"
 	"github.com/mitchellh/mapstructure"
-	doctor "github.com/pb33f/doctor/model"
+	doctorModel "github.com/pb33f/doctor/model"
 	"github.com/pb33f/libopenapi"
 	"github.com/pb33f/libopenapi/datamodel"
 	"github.com/pb33f/libopenapi/index"
 	"github.com/pb33f/libopenapi/utils"
-	"github.com/pterm/pterm"
-	"gopkg.in/yaml.v3"
+	"go.yaml.in/yaml/v4"
 )
 
 type ruleContext struct {
@@ -38,50 +39,89 @@ type ruleContext struct {
 	specNodeUnresolved *yaml.Node
 	builtinFunctions   functions.Functions
 	ruleResults        *[]model.RuleFunctionResult
+	ignoredResults     *[]model.RuleFunctionResult
+	fixedResults       *[]model.RuleFunctionResult
 	errors             *[]error
 	index              *index.SpecIndex
 	specInfo           *datamodel.SpecInfo
 	customFunctions    map[string]model.RuleFunction
+	autoFixFunctions   map[string]model.AutoFixFunction
 	panicFunc          func(p any)
 	silenceLogs        bool
 	document           libopenapi.Document
-	drDocument         *doctor.DrDocument
+	drDocument         *doctorModel.DrDocument
 	skipDocumentCheck  bool
 	logger             *slog.Logger
+	nodeLookupTimeout  time.Duration
+	applyAutoFixes     bool
 }
 
 // RuleSetExecution is an instruction set for executing a ruleset. It's a convenience structure to allow the signature
 // of ApplyRulesToRuleSet to change, without a huge refactor. The ApplyRulesToRuleSet function only returns a single error also.
 type RuleSetExecution struct {
-	RuleSet           *rulesets.RuleSet             // The RuleSet in which to apply
-	SpecFileName      string                        // The name of the specification file, used to correctly label location
-	Spec              []byte                        // The raw bytes of the OpenAPI specification.
-	SpecInfo          *datamodel.SpecInfo           // Pre-parsed spec-info.
-	CustomFunctions   map[string]model.RuleFunction // custom functions loaded from plugin.
-	PanicFunction     func(p any)                   // In case of emergency, do this thing here.
-	SilenceLogs       bool                          // Prevent any warnings about rules/rule-sets being printed.
-	Base              string                        // The base path or URL of the specification, used for resolving relative or remote paths.
-	AllowLookup       bool                          // Allow remote lookup of files or links
-	Document          libopenapi.Document           // a ready to render model.
-	DrDocument        *doctor.DrDocument            // a high level, more powerful model, powered by the doctor.
-	SkipDocumentCheck bool                          // Skip the document check, useful for fragments and non openapi specs.
-	Logger            *slog.Logger                  // A custom logger.
-	Timeout           time.Duration                 // The timeout for each rule to run, prevents run-away rules, default is five seconds.
+	RuleSet                         *rulesets.RuleSet             // The RuleSet in which to apply
+	SpecFileName                    string                        // The path of the specification file, used to correctly label location
+	Spec                            []byte                        // The raw bytes of the OpenAPI specification.
+	SpecInfo                        *datamodel.SpecInfo           // Pre-parsed spec-info.
+	IndexUnresolved                 *index.SpecIndex              // The unresolved index, even if a file is not an OpenAPI spec, it's still indexed.
+	IndexResolved                   *index.SpecIndex              // The resolved index, like the unresolved one, but with references resolved.
+	CustomFunctions                 map[string]model.RuleFunction // custom functions loaded from plugin.
+	AutoFixFunctions                map[string]model.AutoFixFunction // auto-fix functions loaded from plugin.
+	PanicFunction                   func(p any)                   // In case of emergency, do this thing here.
+	SilenceLogs                     bool                          // Prevent any warnings about rules/rule-sets being printed.
+	Base                            string                        // The base path or URL of the specification, used for resolving relative or remote paths.
+	AllowLookup                     bool                          // Allow remote lookup of files or links
+	Document                        libopenapi.Document           // a ready to render model.
+	DrDocument                      *doctorModel.DrDocument       // a high level, more powerful model, powered by the doctorModel.
+	SkipDocumentCheck               bool                          // Skip the document check, useful for fragments and non openapi specs.
+	Logger                          *slog.Logger                  // A custom logger.
+	Timeout                         time.Duration                 // The timeout for each rule to run, prevents run-away rules, default is five seconds.
+	NodeLookupTimeout               time.Duration                 // The timeout for each node yaml path lookup, prevents any endless loops, default is 500ms (https://github.com/daveshanley/vacuum/issues/502)
+	BuildGraph                      bool                          // Build a graph of the document, powered by the doctorModel. (default is false)
+	RenderChanges                   bool                          // Not used by vacuum, used by the openapi doctor (defaults to false).
+	BuildDeepGraph                  bool                          // Build a deep graph of the document, all paths in the graph will be followed, no caching on schemas. (default is false). Required when using ignore files as an object can be referenced in multiple places.
+	ExtractReferencesSequentially   bool                          // Extract references sequentially, defaults to false, can be slow.
+	ExtractReferencesFromExtensions bool                          // Extract references from extension objects (x-), this may pull in all kinds of non-parsable files in.
+	ApplyAutoFixes                  bool                          // Apply auto-fixes for rules that support it
+	CanonicalDocument               *yaml.Node                    // The single source of truth for all modifications
 
 	// https://pb33f.io/libopenapi/circular-references/#circular-reference-results
 	IgnoreCircularArrayRef       bool // Ignore array circular references
 	IgnoreCircularPolymorphicRef bool // Ignore polymorphic circular references
+
+	// not generally used.
+	StorageRoot string // The root path for storage, used for storing files upstream by the doctorModel. You probably don't need this.
+	RolodexFS   fs.FS  // supply a custom local filesystem to be used by the rolodex, useful if you need fine grained control over local file references.
+
+	// HTTP client configuration for TLS/certificate support
+	HTTPClientConfig vacuumUtils.HTTPClientConfig // Configuration for custom HTTP client with certificate support
+}
+
+// buildLocationString efficiently builds a location string in format "line:column"
+// This replaces fmt.Sprintf("%d:%d", line, column) which was a performance bottleneck
+func buildLocationString(line, column int) string {
+	// Pre-allocate with reasonable capacity for most line numbers
+	var builder strings.Builder
+	builder.Grow(12) // Should handle most line:column combinations efficiently
+	builder.WriteString(strconv.Itoa(line))
+	builder.WriteByte(':')
+	builder.WriteString(strconv.Itoa(column))
+	return builder.String()
 }
 
 // RuleSetExecutionResult returns the results of running the ruleset against the supplied spec.
 type RuleSetExecutionResult struct {
-	RuleSetExecution *RuleSetExecution          // The execution struct that was used invoking the result.
-	Results          []model.RuleFunctionResult // The results of the execution.
-	Index            *index.SpecIndex           // The index that was created from the specification, used by the rules.
-	SpecInfo         *datamodel.SpecInfo        // A reference to the SpecInfo object, used by all the rules.
-	Errors           []error                    // Any errors that were returned.
-	FilesProcessed   int                        // number of files extracted by the rolodex
-	FileSize         int64                      // total filesize loaded by the rolodex
+	RuleSetExecution *RuleSetExecution                // The execution struct that was used invoking the result.
+	Results          []model.RuleFunctionResult       // The results of the execution.
+	IgnoredResults   []model.RuleFunctionResult       // Results that were ignored due to inline ignore directives.
+	FixedResults     []model.RuleFunctionResult       // Results that were automatically fixed.
+	Index            *index.SpecIndex                 // The index that was created from the specification, used by the rules.
+	SpecInfo         *datamodel.SpecInfo              // A reference to the SpecInfo object, used by all the rules.
+	Errors           []error                          // Any errors that were returned.
+	FilesProcessed   int                              // number of files extracted by the rolodex
+	FileSize         int64                            // total filesize loaded by the rolodex
+	DocumentConfig   *datamodel.DocumentConfiguration // The document configuration used to create the document.
+	ModifiedSpec     []byte                           // The spec with autofix changes applied (if any fixes were made).
 }
 
 // todo: move copy into virtual file system or some kind of map.
@@ -98,24 +138,27 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 	now := time.Now()
 	builtinFunctions := functions.MapBuiltinFunctions()
 	var ruleResults []model.RuleFunctionResult
+	var ignoredResults []model.RuleFunctionResult
+	var fixedResults []model.RuleFunctionResult
 	var ruleWaitGroup sync.WaitGroup
 	if execution.RuleSet != nil && execution.RuleSet.Rules != nil {
 		ruleWaitGroup.Add(len(execution.RuleSet.Rules))
 	}
 
-	var specResolved *yaml.Node
-	var specUnresolved *yaml.Node
-
 	// create new configurations
 	indexConfig := index.CreateClosedAPIIndexConfig()
+	indexConfig.ExcludeExtensionRefs = true // disable references in extensions being extracted by default
+	indexConfig.SpecFilePath = execution.SpecFileName
 	indexConfigUnresolved := index.CreateClosedAPIIndexConfig()
+	indexConfigUnresolved.SpecFilePath = execution.SpecFileName
 
 	// avoid building the index, we don't need it to run yet.
 	indexConfig.AvoidBuildIndex = true
-	//indexConfig.AvoidCircularReferenceCheck = true
 
 	docConfig := datamodel.NewDocumentConfiguration()
-	//docConfig.SkipCircularReferenceCheck = true
+	docConfig.SpecFilePath = execution.SpecFileName
+	docConfig.LocalFS = execution.RolodexFS
+	docConfig.RemoteFS = execution.RolodexFS
 
 	if execution.IgnoreCircularArrayRef {
 		docConfig.IgnoreArrayCircularReferences = true
@@ -124,6 +167,10 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 	if execution.IgnoreCircularPolymorphicRef {
 		docConfig.IgnorePolymorphicCircularReferences = true
 	}
+	if !execution.ExtractReferencesFromExtensions {
+		indexConfig.ExcludeExtensionRefs = true
+		docConfig.ExcludeExtensionRefs = true
+	}
 
 	// add new pretty logger.
 	if execution.Logger == nil {
@@ -134,20 +181,8 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 				Level: slog.LevelError,
 			}))
 		} else {
-			handler := pterm.NewSlogHandler(&pterm.Logger{
-				Formatter: pterm.LogFormatterColorful,
-				Writer:    os.Stdout,
-				Level:     pterm.LogLevelError,
-				ShowTime:  false,
-				MaxWidth:  280,
-				KeyStyles: map[string]pterm.Style{
-					"error":  *pterm.NewStyle(pterm.FgRed, pterm.Bold),
-					"err":    *pterm.NewStyle(pterm.FgRed, pterm.Bold),
-					"caller": *pterm.NewStyle(pterm.FgGray, pterm.Bold),
-				},
-			})
-			logger = slog.New(handler)
-			pterm.DefaultLogger.Level = pterm.LogLevelError
+			// use simple logger that discards output for internal motor operations
+			logger = slog.New(slog.NewTextHandler(io.Discard, nil))
 		}
 		docConfig.Logger = logger
 		indexConfig.Logger = logger
@@ -159,6 +194,17 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 
 	indexConfig.Logger.Debug("applying rules to rule set")
 
+	// Configure custom HTTP client if TLS/certificate options are provided
+	if vacuumUtils.ShouldUseCustomHTTPClient(execution.HTTPClientConfig) {
+		httpClient, httpErr := vacuumUtils.CreateCustomHTTPClient(execution.HTTPClientConfig)
+		if httpErr != nil {
+			return &RuleSetExecutionResult{Errors: []error{fmt.Errorf("failed to create custom HTTP client: %w", httpErr)}}
+		}
+
+		// Set the custom RemoteURLHandler for libopenapi
+		docConfig.RemoteURLHandler = vacuumUtils.CreateRemoteURLHandler(httpClient)
+	}
+
 	if execution.Base != "" {
 
 		// check if this is a URL or not
@@ -187,9 +233,23 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 	}
 
 	if execution.AllowLookup {
-		if indexConfig.BasePath == "" {
-			indexConfig.BasePath = "."
-			docConfig.BasePath = "."
+		if indexConfig.BasePath == "" && indexConfig.BaseURL == nil {
+			// Use the directory of the spec file as the base path if available
+			// This ensures relative references in the spec work correctly
+			if execution.SpecFileName != "" {
+				specDir := filepath.Dir(execution.SpecFileName)
+				if specDir != "" && specDir != "." {
+					indexConfig.BasePath = specDir
+					docConfig.BasePath = specDir
+					indexConfigUnresolved.BasePath = specDir
+				} else {
+					indexConfig.BasePath = "."
+					docConfig.BasePath = "."
+				}
+			} else {
+				indexConfig.BasePath = "."
+				docConfig.BasePath = "."
+			}
 		}
 		indexConfig.AllowFileLookup = true
 		indexConfigUnresolved.AllowFileLookup = true
@@ -199,6 +259,10 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 		docConfig.AllowFileReferences = true
 	}
 
+	if execution.ExtractReferencesSequentially {
+		docConfig.ExtractRefsSequentially = true
+	}
+
 	if execution.SkipDocumentCheck {
 		docConfig.BypassDocumentCheck = true
 	}
@@ -213,7 +277,8 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 	indexConfig.Logger.Debug("building documents")
 	nowDocs := time.Now()
 
-	var specInfo, specInfoUnresolved *datamodel.SpecInfo
+	var specInfo *datamodel.SpecInfo
+	var specInfoUnresolved *datamodel.SpecInfo
 	if docResolved == nil {
 		var err error
 		// create a new document.
@@ -272,12 +337,59 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 	indexConfig.Logger.Debug("building docs completed", "ms", timeTaken)
 
 	// build model
-	var resolvedModelErrors []error
+	var resolvedModelErrors error
 	var indexResolved *index.SpecIndex
 	var indexUnresolved *index.SpecIndex
 
 	version := docResolved.GetVersion()
 
+	// When skip-check is enabled, the document version might not be detected
+	// but we still need to know if it's OAS2 or OAS3 for rule filtering
+	if version == "" && execution.SkipDocumentCheck && specInfo != nil {
+		// Try to detect the version from the spec directly
+		if specInfo.SpecType != "" {
+			if strings.HasPrefix(strings.ToLower(specInfo.SpecType), "swagger") {
+				version = "2.0"
+				specInfo.SpecFormat = model.OAS2
+				if specInfoUnresolved != nil {
+					specInfoUnresolved.SpecFormat = model.OAS2
+				}
+			} else if strings.HasPrefix(strings.ToLower(specInfo.SpecType), "openapi") {
+				// Try to get more specific version from SpecVersion field
+				if specInfo.Version != "" {
+					version = specInfo.Version
+					if strings.HasPrefix(specInfo.Version, "3.2") {
+						// Note: OAS 3.2 is set in specInfo when skip-check is NOT used,
+						// but libopenapi's BuildV3Model doesn't handle OAS32 yet, so we
+						// temporarily map it to OAS3 for model building while preserving
+						// the actual format for rule filtering
+						specInfo.SpecFormat = model.OAS32
+						if specInfoUnresolved != nil {
+							specInfoUnresolved.SpecFormat = model.OAS32
+						}
+					} else if strings.HasPrefix(specInfo.Version, "3.1") {
+						specInfo.SpecFormat = model.OAS31
+						if specInfoUnresolved != nil {
+							specInfoUnresolved.SpecFormat = model.OAS31
+						}
+					} else {
+						specInfo.SpecFormat = model.OAS3
+						if specInfoUnresolved != nil {
+							specInfoUnresolved.SpecFormat = model.OAS3
+						}
+					}
+				} else {
+					// Default to OAS3 if we know it's OpenAPI but not the specific version
+					version = "3.0"
+					specInfo.SpecFormat = model.OAS3
+					if specInfoUnresolved != nil {
+						specInfoUnresolved.SpecFormat = model.OAS3
+					}
+				}
+			}
+		}
+	}
+
 	var resolvingErrors []*index.ResolvingError
 	var circularReferences []*index.CircularReferenceResult
 
@@ -289,7 +401,7 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 	var v3DocumentModel *v3.Document
 	//var v2DocumentModel *v2.Swagger
 
-	var drDocument *doctor.DrDocument
+	var drDocument *doctorModel.DrDocument
 
 	if version != "" {
 		switch version[0] {
@@ -303,11 +415,12 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 			indexResolved = rolodexResolved.GetRootIndex()
 			indexUnresolved = rolodexUnresolved.GetRootIndex()
 
+			// Set the canonical document to the unresolved spec for autofix modifications
+			execution.CanonicalDocument = rolodexUnresolved.GetRootIndex().GetRootNode()
+
 			// we only resolve one.
 			rolodexResolved.Resolve()
 
-			specResolved = rolodexResolved.GetRootIndex().GetRootNode()
-			specUnresolved = rolodexUnresolved.GetRootIndex().GetRootNode()
 
 			if rolodexResolved != nil && rolodexResolved.GetRootIndex() != nil {
 				resolvingErrors = rolodexResolved.GetRootIndex().GetResolver().GetResolvingErrors()
@@ -327,9 +440,14 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 			indexConfig.Logger.Debug("built resolved model", "ms", then)
 
 			now = time.Now()
-			mod, _ = docUnresolved.BuildV3Model()
+			var errs error
+			mod, errs = docUnresolved.BuildV3Model()
 			if mod != nil {
 				v3DocumentModel = &mod.Model
+			} else {
+				if execution.Logger != nil {
+					execution.Logger.Error("unable to build unresolved model", "errors", errs)
+				}
 			}
 			rolodexUnresolved = docUnresolved.GetRolodex()
 
@@ -342,7 +460,27 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 			wg := conc.WaitGroup{}
 			wg.Go(func() {
 				if v3DocumentModel != nil {
-					drDocument = doctor.NewDrDocument(mod)
+					var drDoc *doctorModel.DrDocument
+					if execution.StorageRoot != "" {
+						mod.Model.GoLow().StorageRoot = execution.StorageRoot
+					}
+
+					buildGraph := false
+					useCache := true
+					if execution.BuildGraph {
+						buildGraph = true
+					}
+					if execution.BuildDeepGraph {
+						useCache = false
+					}
+					drDoc = doctorModel.NewDrDocumentWithConfig(mod, &doctorModel.DrConfig{
+						BuildGraph:     buildGraph,
+						UseSchemaCache: useCache,
+						RenderChanges:  execution.RenderChanges,
+					})
+
+					execution.DrDocument = drDoc
+					drDocument = drDoc
 				}
 			})
 			wg.Go(func() {
@@ -353,8 +491,6 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 				indexConfig.Logger.Debug("resolved model", "ms", resolvedTaken)
 			})
 			wg.Wait()
-			specResolved = rolodexResolved.GetRootIndex().GetRootNode()
-			specUnresolved = rolodexUnresolved.GetRootIndex().GetRootNode()
 
 			if rolodexResolved != nil && rolodexResolved.GetRootIndex() != nil {
 				//resolvingErrors = rolodexResolved.GetRootIndex().GetResolver().GetResolvingErrors()
@@ -371,27 +507,29 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 
 		wg.Go(func() {
 			// create an index for the unresolved spec.
-			rolodexResolved, _ = BuildRolodexFromIndexConfig(&resRoloConfig)
+			rolodexResolved, _ = BuildRolodexFromIndexConfig(&resRoloConfig, execution.RolodexFS)
 			rolodexResolved.SetRootNode(resRoloConfig.SpecInfo.RootNode)
 
-			_ = rolodexResolved.IndexTheRolodex()
+			_ = rolodexResolved.IndexTheRolodex(context.Background())
 			rolodexResolved.Resolve()
 		})
 
 		wg.Go(func() {
 			unResInfo, _ := datamodel.ExtractSpecInfo(*specInfo.SpecBytes)
-			rolodexUnresolved, _ = BuildRolodexFromIndexConfig(&unresRoloConfig)
-			rolodexUnresolved.SetRootNode(unResInfo.RootNode)
-
-			_ = rolodexUnresolved.IndexTheRolodex()
+			rolodexUnresolved, _ = BuildRolodexFromIndexConfig(&unresRoloConfig, execution.RolodexFS)
+			if unResInfo != nil {
+				rolodexUnresolved.SetRootNode(unResInfo.RootNode)
+				_ = rolodexUnresolved.IndexTheRolodex(context.Background())
+			}
 		})
 		wg.Wait()
 
 		indexResolved = rolodexResolved.GetRootIndex()
 		indexUnresolved = rolodexUnresolved.GetRootIndex()
 
-		specResolved = rolodexResolved.GetRootNode()
-		specUnresolved = rolodexUnresolved.GetRootNode()
+
+		// Set the canonical document to the unresolved spec for autofix modifications
+		execution.CanonicalDocument = rolodexUnresolved.GetRootNode()
 
 		if rolodexResolved != nil && rolodexResolved.GetRootIndex() != nil {
 			resolvingErrors = rolodexResolved.GetRootIndex().GetResolver().GetResolvingErrors()
@@ -402,9 +540,12 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 	then := time.Since(nowModel).Milliseconds()
 	indexConfig.Logger.Debug("built model", "ms", then)
 
-	for i := range resolvedModelErrors {
+	execution.IndexResolved = indexResolved
+	execution.IndexUnresolved = indexUnresolved
+	r := utils.UnwrapErrors(resolvedModelErrors)
+	for i := range r {
 		var m *index.ResolvingError
-		if errors.As(resolvedModelErrors[i], &m) {
+		if errors.As(r[i], &m) {
 			resolvingErrors = append(resolvingErrors, m)
 		}
 	}
@@ -492,6 +633,16 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 			Message:   er.Error(),
 			Path:      er.Path,
 		}
+		if res.StartNode == nil {
+			res.StartNode = utils.CreateStringNode("")
+			res.StartNode.Line = 1
+			res.StartNode.Column = 1
+		}
+		if res.EndNode == nil {
+			res.EndNode = utils.CreateStringNode("")
+			res.EndNode.Line = 1
+			res.EndNode.Column = 1
+		}
 		ruleResults = append(ruleResults, res)
 	}
 
@@ -502,9 +653,19 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 			Rule:      circularRefRule,
 			StartNode: cr.ParentNode,
 			EndNode:   vacuumUtils.BuildEndNode(cr.ParentNode),
-			Message:   fmt.Sprintf("circular reference detected from %s", cr.Start.Definition),
+			Message:   "circular reference detected from " + cr.Start.Definition,
 			Path:      cr.GenerateJourneyPath(),
 		}
+		if res.StartNode == nil {
+			res.StartNode = utils.CreateStringNode("")
+			res.StartNode.Line = 1
+			res.StartNode.Column = 1
+		}
+		if res.EndNode == nil {
+			res.EndNode = utils.CreateStringNode("")
+			res.EndNode.Line = 1
+			res.EndNode.Column = 1
+		}
 		ruleResults = append(ruleResults, res)
 	}
 
@@ -520,6 +681,17 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 				Message:   idxError.Error(),
 				Path:      idxError.Path,
 			}
+			if res.StartNode == nil {
+				res.StartNode = utils.CreateStringNode("")
+				res.StartNode.Line = 1
+				res.StartNode.Column = 1
+			}
+			if res.EndNode == nil {
+				res.EndNode = utils.CreateStringNode("")
+				res.EndNode.Line = 1
+				res.EndNode.Column = 1
+			}
+
 			ruleResults = append(ruleResults, res)
 		}
 	}
@@ -527,6 +699,13 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 	// run all rules.
 	var errs []error
 
+	// Ensure CanonicalDocument is set as a fallback
+	if execution.CanonicalDocument == nil {
+		if indexUnresolved != nil && indexUnresolved.GetRootNode() != nil {
+			execution.CanonicalDocument = indexUnresolved.GetRootNode()
+		}
+	}
+
 	// add dr document build errors to the results.
 	if drDocument != nil {
 		for _, er := range drDocument.BuildErrors {
@@ -534,7 +713,7 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 				RuleId:    "schema-build-failure",
 				Rule:      schemaBuildRule,
 				StartNode: er.SchemaProxy.GoLow().GetKeyNode(),
-				EndNode:   vacuumUtils.BuildEndNode(er.SchemaProxy.GoLow().GetValueNode()),
+				EndNode:   vacuumUtils.BuildEndNode(er.SchemaProxy.GoLow().GetKeyNode()),
 				Message:   er.Error.Error(),
 				Path:      er.DrSchemaProxy.GenerateJSONPath(),
 			}
@@ -561,43 +740,52 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 		done := make(chan bool)
 		indexConfig.Logger.Debug("running rules", "total", totalRules)
 		now = time.Now()
+
+		// if there are no time outs, set them to defaults
+		if execution.Timeout <= 0 {
+			execution.Timeout = time.Second * 5
+		}
+		if execution.NodeLookupTimeout <= 0 {
+			execution.NodeLookupTimeout = time.Millisecond * 500
+		}
+
 		for _, rule := range execution.RuleSet.Rules {
 
 			go func(rule *model.Rule, done chan bool) {
 
-				ruleSpec := specResolved
 				ruleIndex := indexResolved
 				info := specInfo
 				if !rule.Resolved {
 					info = specInfoUnresolved
-					ruleSpec = specUnresolved
 					ruleIndex = indexUnresolved
 				}
+
 				// this list of things is most likely going to grow a bit, so we use a nice clean message design.
 				ctx := ruleContext{
 					rule:               rule,
-					specNode:           ruleSpec,
-					specNodeUnresolved: specUnresolved,
+					specNode:           execution.CanonicalDocument,
+					specNodeUnresolved: execution.CanonicalDocument,
 					builtinFunctions:   builtinFunctions,
 					ruleResults:        &ruleResults,
+					ignoredResults:     &ignoredResults,
+					fixedResults:       &fixedResults,
 					errors:             &errs,
 					specInfo:           info,
 					index:              ruleIndex,
 					document:           docUnresolved,
 					drDocument:         drDocument,
 					customFunctions:    execution.CustomFunctions,
+					autoFixFunctions:   execution.AutoFixFunctions,
 					silenceLogs:        execution.SilenceLogs,
 					skipDocumentCheck:  execution.SkipDocumentCheck,
 					logger:             docConfig.Logger,
+					nodeLookupTimeout:  execution.NodeLookupTimeout,
+					applyAutoFixes:     execution.ApplyAutoFixes,
 				}
 				if execution.PanicFunction != nil {
 					ctx.panicFunc = execution.PanicFunction
 				}
 
-				if execution.Timeout <= 0 {
-					execution.Timeout = time.Second * 5 // default
-				}
-
 				timeoutCtx, ruleCancel := context.WithTimeout(context.Background(), execution.Timeout)
 				defer ruleCancel()
 				doneChan := make(chan bool)
@@ -636,19 +824,47 @@ func ApplyRulesToRuleSet(execution *RuleSetExecution) *RuleSetExecutionResult {
 	then = time.Since(now).Milliseconds()
 	indexConfig.Logger.Debug("applied all rules and completed", "ms", then)
 
+	// Marshal modified spec if fixes were applied
+	var modifiedSpec []byte
+	if len(fixedResults) > 0 {
+		var err error
+		// Use the canonical document which contains all modifications
+		modifiedSpec, err = yaml.Marshal(execution.CanonicalDocument)
+		if err != nil {
+			indexConfig.Logger.Warn("failed to marshal modified spec", "error", err)
+		} else {
+			indexConfig.Logger.Debug("ModifiedSpec marshaled from CanonicalDocument")
+		}
+	}
+
 	return &RuleSetExecutionResult{
 		RuleSetExecution: execution,
 		Results:          ruleResults,
+		IgnoredResults:   ignoredResults,
+		FixedResults:     fixedResults,
 		Index:            indexResolved,
 		SpecInfo:         specInfo,
 		Errors:           errs,
 		FilesProcessed:   filesProcessed,
 		FileSize:         fileSize,
+		DocumentConfig:   docConfig,
+		ModifiedSpec:     modifiedSpec,
 	}
 }
 
 func runRule(ctx ruleContext, doneChan chan bool) {
 
+	// Check for missing auto-fix functions when --fix is enabled
+	if ctx.applyAutoFixes && ctx.rule.AutoFixFunction != "" {
+		if _, exists := ctx.autoFixFunctions[ctx.rule.AutoFixFunction]; !exists {
+			if !ctx.silenceLogs {
+				ctx.logger.Warn("Rule uses unknown auto-fix function", 
+					"ruleId", ctx.rule.Id, 
+					"autoFixFunction", ctx.rule.AutoFixFunction)
+			}
+		}
+	}
+
 	if ctx.panicFunc != nil {
 		defer func() {
 			if r := recover(); r != nil {
@@ -675,7 +891,7 @@ func runRule(ctx ruleContext, doneChan chan bool) {
 	}
 
 	findNodes := func(node *yaml.Node, path string, errChan chan error, nodesChan chan []*yaml.Node) {
-		nodes, err := utils.FindNodesWithoutDeserializing(node, path)
+		nodes, err := utils.FindNodesWithoutDeserializingWithTimeout(node, path, ctx.nodeLookupTimeout)
 		if err != nil {
 			errChan <- err
 		}
@@ -732,7 +948,9 @@ func runRule(ctx ruleContext, doneChan chan bool) {
 		}
 
 		if err != nil {
+			lock.Lock()
 			*ctx.errors = append(*ctx.errors, err)
+			lock.Unlock()
 			doneChan <- true
 			return
 		}
@@ -794,7 +1012,7 @@ func buildResults(ctx ruleContext, ruleAction model.RuleAction, nodes []*yaml.No
 
 		if !ctx.skipDocumentCheck && ctx.specInfo.SpecFormat == "" && ctx.specInfo.Version == "" {
 			if !ctx.silenceLogs {
-				pterm.Warning.Printf("Specification version not detected, cannot apply rule `%s`\n", ctx.rule.Id)
+				fmt.Printf("⚠️  Specification version not detected, cannot apply rule `%s`\n", ctx.rule.Id)
 			}
 			return ctx.ruleResults
 		}
@@ -804,12 +1022,23 @@ func buildResults(ctx ruleContext, ruleAction model.RuleAction, nodes []*yaml.No
 		if !res {
 			for _, e := range errs {
 				lock.Lock()
-				*ctx.ruleResults = append(*ctx.ruleResults, model.RuleFunctionResult{Message: e})
+				*ctx.ruleResults = append(*ctx.ruleResults, model.RuleFunctionResult{
+					Message:      e,
+					Rule:         ctx.rule,
+					StartNode:    &yaml.Node{},
+					EndNode:      &yaml.Node{},
+					RuleId:       ctx.rule.Id,
+					RuleSeverity: ctx.rule.Severity,
+					Path:         fmt.Sprint(ctx.rule.Given),
+				})
 				lock.Unlock()
 			}
 		} else {
+			// Filter out ignore nodes to prevent them from being processed by other rules
+			filteredNodes := filterIgnoreNodes(nodes)
+
 			// iterate through nodes and supply them one at a time so we don't pollute each run
-			for _, node := range nodes {
+			for _, node := range filteredNodes {
 
 				// if this rule is designed for a different version, skip it.
 				if len(ctx.rule.Formats) > 0 {
@@ -824,20 +1053,154 @@ func buildResults(ctx ruleContext, ruleAction model.RuleAction, nodes []*yaml.No
 					}
 				}
 
+				if checkInlineIgnore(node, ctx.rule.Id) {
+					ignoredResult := model.RuleFunctionResult{
+						Message:      "Rule ignored due to inline ignore directive",
+						RuleId:       ctx.rule.Id,
+						RuleSeverity: ctx.rule.Severity,
+						Rule:         ctx.rule,
+						StartNode:    node,
+						EndNode:      node,
+					}
+					
+					lock.Lock()
+					*ctx.ignoredResults = append(*ctx.ignoredResults, ignoredResult)
+					lock.Unlock()
+					continue
+				}
+
 				runRuleResults := ruleFunction.RunRule([]*yaml.Node{node}, rfc)
 
-				// because this function is running in multiple threads, we need to sync access to the final result
-				// list, otherwise things can get a bit random.
-				lock.Lock()
-				*ctx.ruleResults = append(*ctx.ruleResults, runRuleResults...)
-				lock.Unlock()
+				// Filter out results that should be ignored due to inline ignore directives
+				var filteredResults []model.RuleFunctionResult
+				for _, result := range runRuleResults {
+					// Check if this result should be ignored based on its path
+					if result.Path != "" && checkInlineIgnoreByPath(ctx.specNode, result.Path, ctx.rule.Id) {
+						ignoredResult := model.RuleFunctionResult{
+							Message:      "Rule ignored due to inline ignore directive",
+							RuleId:       ctx.rule.Id,
+							RuleSeverity: ctx.rule.Severity,
+							Rule:         ctx.rule,
+							StartNode:    result.StartNode,
+							EndNode:      result.EndNode,
+							Path:         result.Path,
+						}
+						lock.Lock()
+						*ctx.ignoredResults = append(*ctx.ignoredResults, ignoredResult)
+						lock.Unlock()
+					} else {
+						filteredResults = append(filteredResults, result)
+					}
+				}
+
+				// Ensure RuleId and RuleSeverity are populated from the rule context
+				// This is necessary for programmatic API usage where these fields might not be set
+				for i := range filteredResults {
+					if filteredResults[i].RuleId == "" {
+						filteredResults[i].RuleId = ctx.rule.Id
+					}
+					if filteredResults[i].RuleSeverity == "" {
+						filteredResults[i].RuleSeverity = ctx.rule.Severity
+					}
+					if filteredResults[i].Rule == nil {
+						filteredResults[i].Rule = ctx.rule
+					}
+				}
+
+				// Apply auto-fix if available and enabled
+				if ctx.applyAutoFixes && ctx.rule.AutoFixFunction != "" {
+					applyAutoFixesToResults(ctx, filteredResults, &rfc)
+				} else {
+					// No autofix - add all results to regular results
+					// because this function is running in multiple threads, we need to sync access to the final result
+					// list, otherwise things can get a bit random.
+					lock.Lock()
+					*ctx.ruleResults = append(*ctx.ruleResults, filteredResults...)
+					lock.Unlock()
+				}
 			}
 
 		}
+	} else {
+		// Function not found - report detailed error
+		if !ctx.silenceLogs {
+			// Build list of available custom functions for debugging
+			var availableCustomFuncs []string
+			if ctx.customFunctions != nil {
+				for funcName := range ctx.customFunctions {
+					availableCustomFuncs = append(availableCustomFuncs, funcName)
+				}
+			}
+
+			if len(availableCustomFuncs) > 0 {
+				fmt.Printf("✗ Rule '%s' uses unknown function '%s'. Available custom functions: %v\n",
+					ctx.rule.Id, ruleAction.Function, availableCustomFuncs)
+			} else {
+				fmt.Printf("✗ Rule '%s' uses unknown function '%s'. No custom functions loaded. Use --functions flag to load custom functions.\n",
+					ctx.rule.Id, ruleAction.Function)
+			}
+		}
+
+		// Add error result to make the missing function visible in reports
+		lock.Lock()
+		*ctx.ruleResults = append(*ctx.ruleResults, model.RuleFunctionResult{
+			Message:      fmt.Sprintf("Unknown function '%s' in rule '%s'", ruleAction.Function, ctx.rule.Id),
+			Rule:         ctx.rule,
+			StartNode:    &yaml.Node{},
+			EndNode:      &yaml.Node{},
+			RuleId:       ctx.rule.Id,
+			RuleSeverity: "error",
+			Path:         fmt.Sprint(ctx.rule.Given),
+		})
+		lock.Unlock()
 	}
 	return ctx.ruleResults
 }
 
+func applyAutoFixesToResults(ctx ruleContext, results []model.RuleFunctionResult, rfc *model.RuleFunctionContext) {
+	lock := sync.Mutex{}
+	
+	for i := range results {
+		// Only attempt auto-fix if there's a violation and we have the node
+		if results[i].StartNode == nil {
+			lock.Lock()
+			*ctx.ruleResults = append(*ctx.ruleResults, results[i])
+			lock.Unlock()
+			continue
+		}
+
+		autoFixFunc, exists := ctx.autoFixFunctions[ctx.rule.AutoFixFunction]
+		if !exists {
+			// Auto-fix function not found - add to regular results
+			lock.Lock()
+			*ctx.ruleResults = append(*ctx.ruleResults, results[i])
+			lock.Unlock()
+			continue
+		}
+
+		_, err := autoFixFunc(results[i].StartNode, ctx.specNode, rfc)
+		if err != nil {
+			// Auto-fix failed - add to regular results
+			if !ctx.silenceLogs {
+				ctx.logger.Warn("Auto-fix failed", "ruleId", ctx.rule.Id, "error", err)
+			}
+			lock.Lock()
+			*ctx.ruleResults = append(*ctx.ruleResults, results[i])
+			lock.Unlock()
+		} else {
+			// Auto-fix succeeded - add to fixed results
+			results[i].AutoFixed = true
+			lock.Lock()
+			*ctx.fixedResults = append(*ctx.fixedResults, results[i])
+			lock.Unlock()
+
+			if !ctx.silenceLogs {
+				ctx.logger.Debug("Auto-fix applied", "ruleId", ctx.rule.Id, "path", results[i].Path)
+			}
+		}
+	}
+}
+
 type seenResult struct {
 	location string
 	message  string
@@ -854,7 +1217,7 @@ func removeDuplicates(results *[]model.RuleFunctionResult, rse *RuleSetExecution
 			if result.StartNode != nil {
 				seen[result.RuleId] = []*seenResult{
 					{
-						fmt.Sprintf("%d:%d", result.StartNode.Line, result.StartNode.Column),
+						buildLocationString(result.StartNode.Line, result.StartNode.Column),
 						result.Message,
 					},
 				}
@@ -878,14 +1241,14 @@ func removeDuplicates(results *[]model.RuleFunctionResult, rse *RuleSetExecution
 					continue
 				}
 
-				if line.location == fmt.Sprintf("%d:%d", result.StartNode.Line, result.StartNode.Column) &&
+				if line.location == buildLocationString(result.StartNode.Line, result.StartNode.Column) &&
 					line.message == result.Message {
 					break stopNowPlease
 				}
 				if result.StartNode != nil {
 					seen[result.RuleId] = []*seenResult{
 						{
-							fmt.Sprintf("%d:%d", result.StartNode.Line, result.StartNode.Column),
+							buildLocationString(result.StartNode.Line, result.StartNode.Column),
 							result.Message,
 						},
 					}
