@@ -10,19 +10,27 @@ import (
 	"errors"
 	"fmt"
 	"log/slog"
+	"net/http"
 	"os"
 	"path/filepath"
 	"strings"
+	"sync"
 	"time"
 
+	"golang.org/x/text/language"
+	"golang.org/x/text/message"
+
 	"github.com/daveshanley/vacuum/model"
 	"github.com/mitchellh/mapstructure"
 	"github.com/pb33f/libopenapi/utils"
-	"github.com/santhosh-tekuri/jsonschema/v5"
+	"github.com/santhosh-tekuri/jsonschema/v6"
 )
 
 //go:embed schemas/ruleset.schema.json
-var rulesetSchema string
+var RulesetSchema string
+
+//go:embed schemas/rule.schema.json
+var RuleSchema string
 
 const (
 	Style                                = "style"
@@ -42,6 +50,7 @@ const (
 	InfoContact                          = "info-contact"
 	InfoDescription                      = "info-description"
 	InfoLicense                          = "info-license"
+	InfoLicenseSPDX                      = "info-license-spdx"
 	LicenseUrl                           = "license-url"
 	OpenAPITagsAlphabetical              = "openapi-tags-alphabetical"
 	OpenAPITags                          = "openapi-tags"
@@ -54,6 +63,7 @@ const (
 	PathNotIncludeQuery                  = "path-not-include-query"
 	TagDescription                       = "tag-description"
 	NoRefSiblings                        = "no-$ref-siblings"
+	Oas3NoRefSiblings                    = "oas3-no-$ref-siblings"
 	Oas3UnusedComponent                  = "oas3-unused-component"
 	Oas2UnusedDefinition                 = "oas2-unused-definition"
 	Oas2APIHost                          = "oas2-api-host"
@@ -72,6 +82,7 @@ const (
 	Oas3ExampleExternalCheck             = "oas3-example-external-check"
 	TypedEnum                            = "typed-enum"
 	DuplicatedEntryInEnum                = "duplicated-entry-in-enum"
+	NullableEnumRule                     = "nullable-enum-contains-null"
 	NoEvalInMarkdown                     = "no-eval-in-markdown"
 	NoScriptTagsInMarkdown               = "no-script-tags-in-markdown"
 	DescriptionDuplication               = "description-duplication"
@@ -82,6 +93,12 @@ const (
 	Oas2Schema                           = "oas2-schema"
 	Oas3Schema                           = "oas3-schema"
 	OasSchemaCheck                       = "oas-schema-check"
+	OasMissingType                       = "oas-missing-type"
+	PathItemReferences                   = "path-item-refs"
+	DuplicatePathsRule                   = "duplicate-paths"
+	UnnecessaryCombinatorRule            = "no-unnecessary-combinator"
+	CamelCasePropertiesRule              = "camel-case-properties"
+	MigrateZallyIgnoreRule               = "migrate-zally-ignore"
 	OwaspNoNumericIDs                    = "owasp-no-numeric-ids"
 	OwaspNoHttpBasic                     = "owasp-no-http-basic"
 	OwaspNoAPIKeysInURL                  = "owasp-no-api-keys-in-url"
@@ -105,12 +122,19 @@ const (
 	OwaspNoAdditionalProperties          = "owasp-no-additionalProperties"
 	OwaspConstrainedAdditionalProperties = "owasp-constrained-additionalProperties"
 	OwaspSecurityHostsHttpsOAS3          = "owasp-security-hosts-https-oas3"
+	PostResponseSuccess                  = "post-response-success"
+	NoRequestBody                        = "no-request-body"
+	VacuumOpenAPI                        = "vacuum:oas"
 	SpectralOpenAPI                      = "spectral:oas"
 	SpectralOwasp                        = "spectral:owasp"
 	VacuumOwasp                          = "vacuum:owasp"
-	SpectralRecommended                  = "recommended"
-	SpectralAll                          = "all"
-	SpectralOff                          = "off"
+	VacuumAllRulesets                    = "vacuum:all"  // Combined OpenAPI + OWASP rules
+	VacuumRecommended                    = "recommended"
+	VacuumAll                            = "all"
+	VacuumOff                            = "off"
+	SpectralRecommended                  = VacuumRecommended
+	SpectralAll                          = VacuumAll
+	SpectralOff                          = SpectralAll
 )
 
 type ruleSetsModel struct {
@@ -132,6 +156,11 @@ type RuleSets interface {
 	// GenerateRuleSetFromSuppliedRuleSet will generate a ready to run ruleset based on a supplied configuration. This
 	// will look for any extensions and apply all rules turned on, turned off and any custom rules.
 	GenerateRuleSetFromSuppliedRuleSet(config *RuleSet) *RuleSet
+	
+	// GenerateRuleSetFromSuppliedRuleSetWithHTTPClient will generate a ready to run ruleset based on a supplied configuration. This
+	// will look for any extensions and apply all rules turned on, turned off and any custom rules.
+	// It accepts an HTTP client for downloading remote rulesets with certificate authentication.
+	GenerateRuleSetFromSuppliedRuleSetWithHTTPClient(config *RuleSet, httpClient *http.Client) *RuleSet
 }
 
 //var rulesetsSingleton *ruleSetsModel
@@ -173,6 +202,10 @@ func (rsm ruleSetsModel) GenerateOpenAPIRecommendedRuleSet() *RuleSet {
 }
 
 func (rsm ruleSetsModel) GenerateRuleSetFromSuppliedRuleSet(ruleset *RuleSet) *RuleSet {
+	return rsm.GenerateRuleSetFromSuppliedRuleSetWithHTTPClient(ruleset, nil)
+}
+
+func (rsm ruleSetsModel) GenerateRuleSetFromSuppliedRuleSetWithHTTPClient(ruleset *RuleSet, httpClient *http.Client) *RuleSet {
 
 	extends := ruleset.GetExtendsValue()
 
@@ -186,17 +219,43 @@ func (rsm ruleSetsModel) GenerateRuleSetFromSuppliedRuleSet(ruleset *RuleSet) *R
 	}
 
 	// default and explicitly recommended
-	if extends[SpectralOpenAPI] == SpectralRecommended || extends[SpectralOpenAPI] == SpectralOpenAPI {
+	if extends[VacuumOpenAPI] == VacuumRecommended || extends[VacuumOpenAPI] == VacuumOpenAPI {
+		rs = rsm.GenerateOpenAPIRecommendedRuleSet()
+	}
+
+	// default and explicitly recommended
+	if extends[SpectralOpenAPI] == VacuumRecommended || extends[SpectralOpenAPI] == SpectralOpenAPI {
 		rs = rsm.GenerateOpenAPIRecommendedRuleSet()
 	}
 
 	// all rules
-	if extends[SpectralOpenAPI] == SpectralAll {
+	if extends[SpectralOpenAPI] == VacuumAll || extends[VacuumOpenAPI] == VacuumAll {
 		rs = rsm.openAPIRuleSet
 	}
 
+	// vacuum:all - combines both OpenAPI and OWASP rules
+	if extends[VacuumAllRulesets] == VacuumAll || extends[VacuumAllRulesets] == VacuumAllRulesets {
+		// Start with OpenAPI rules
+		rs = rsm.openAPIRuleSet
+		// Add all OWASP rules
+		for ruleName, rule := range GetAllOWASPRules() {
+			rs.Rules[ruleName] = rule
+		}
+		rs.DocumentationURI = "https://quobix.com/vacuum/rulesets/all-combined"
+		rs.Description = "All OpenAPI and OWASP rules combined"
+	}
+
+	// vacuum:all with off - start with empty ruleset
+	if extends[VacuumAllRulesets] == VacuumOff {
+		if rs.DocumentationURI == "" {
+			rs.DocumentationURI = "https://quobix.com/vacuum/rulesets/no-rules"
+		}
+		rs.Rules = make(map[string]*model.Rule)
+		rs.Description = fmt.Sprintf("All disabled ruleset, processing %d supplied rules", len(rs.RuleDefinitions))
+	}
+
 	// no rules!
-	if extends[SpectralOpenAPI] == SpectralOff {
+	if extends[SpectralOpenAPI] == VacuumOff || extends[VacuumOpenAPI] == VacuumOff {
 		if rs.DocumentationURI == "" {
 			rs.DocumentationURI = "https://quobix.com/vacuum/rulesets/no-rules"
 		}
@@ -214,14 +273,14 @@ func (rsm ruleSetsModel) GenerateRuleSetFromSuppliedRuleSet(ruleset *RuleSet) *R
 	}
 
 	// owasp rules with spectral and vacuum namespace
-	if extends[SpectralOwasp] == SpectralAll || extends[VacuumOwasp] == SpectralAll {
+	if extends[SpectralOwasp] == VacuumAll || extends[VacuumOwasp] == VacuumAll {
 		for ruleName, rule := range GetAllOWASPRules() {
 			rs.Rules[ruleName] = rule
 		}
 	}
 
 	// owasp rules with spectral and vacuum namespace (recommended)
-	if extends[SpectralOwasp] == SpectralRecommended || extends[VacuumOwasp] == SpectralRecommended {
+	if extends[SpectralOwasp] == VacuumRecommended || extends[VacuumOwasp] == VacuumRecommended {
 		for ruleName, rule := range GetRecommendedOWASPRules() {
 			rs.Rules[ruleName] = rule
 		}
@@ -256,7 +315,7 @@ func (rsm ruleSetsModel) GenerateRuleSetFromSuppliedRuleSet(ruleset *RuleSet) *R
 					if strings.HasPrefix(k, "http") {
 						remote = true
 					}
-					SniffOutAllExternalRules(ctx, &rsm, k, nil, rs, remote)
+					SniffOutAllExternalRules(ctx, &rsm, k, nil, rs, remote, httpClient)
 				}
 			}
 			doneChan <- true
@@ -274,6 +333,7 @@ func (rsm ruleSetsModel) GenerateRuleSetFromSuppliedRuleSet(ruleset *RuleSet) *R
 
 	// now all the base rules are in, let's run through the raw definitions and decide
 	// what we need to add, enable, disable, replace or change severity on.
+	rs.mutex.Lock()
 	for k, v := range rs.RuleDefinitions {
 
 		// let's try to cast to a string first (enable/disable/severity)
@@ -291,7 +351,7 @@ func (rsm ruleSetsModel) GenerateRuleSetFromSuppliedRuleSet(ruleset *RuleSet) *R
 			switch evalStr {
 			case model.SeverityError, model.SeverityWarn, model.SeverityInfo, model.SeverityHint:
 				rs.Rules[k].Severity = evalStr
-			case SpectralOff:
+			case VacuumOff:
 				delete(rs.Rules, k) // remove it completely
 			}
 		}
@@ -300,11 +360,26 @@ func (rsm ruleSetsModel) GenerateRuleSetFromSuppliedRuleSet(ruleset *RuleSet) *R
 		// otherwise it means delete it
 		if eval, ok := v.(bool); ok {
 			if eval {
-				if rsm.openAPIRuleSet.Rules[k] == nil {
-					rsm.logger.Warn("Rule does not exist, ignoring it", "rule", k)
-					continue
+				// First check if it's in the OpenAPI ruleset
+				if rsm.openAPIRuleSet.Rules[k] != nil {
+					rs.Rules[k] = rsm.openAPIRuleSet.Rules[k]
+				} else {
+					// Check if it's an OWASP rule when vacuum:all or vacuum:owasp is used
+					if extends[VacuumAllRulesets] == VacuumOff || extends[VacuumAllRulesets] == VacuumAll || extends[VacuumAllRulesets] == VacuumAllRulesets ||
+						extends[VacuumOwasp] == VacuumOff || extends[VacuumOwasp] == VacuumAll || extends[VacuumOwasp] == VacuumRecommended ||
+						extends[SpectralOwasp] == VacuumOff || extends[SpectralOwasp] == VacuumAll || extends[SpectralOwasp] == VacuumRecommended {
+						allOWASPRules := GetAllOWASPRules()
+						if allOWASPRules[k] != nil {
+							rs.Rules[k] = allOWASPRules[k]
+						} else {
+							rsm.logger.Warn("Rule does not exist, ignoring it", "rule", k)
+							continue
+						}
+					} else {
+						rsm.logger.Warn("Rule does not exist, ignoring it", "rule", k)
+						continue
+					}
 				}
-				rs.Rules[k] = rsm.openAPIRuleSet.Rules[k]
 			} else {
 				delete(rs.Rules, k) // remove it completely
 			}
@@ -334,9 +409,16 @@ func (rsm ruleSetsModel) GenerateRuleSetFromSuppliedRuleSet(ruleset *RuleSet) *R
 			} else {
 				if model.RuleCategories[rc.Id] != nil {
 					nr.RuleCategory = model.RuleCategories[rc.Id]
+				} else {
+					model.RuleCategories[rc.Id] = &rc
+					nr.RuleCategory = model.RuleCategories[rc.Id]
 				}
 			}
 
+			if nr.RuleCategory == nil && rs.Rules[k].RuleCategory != nil {
+				nr.RuleCategory = rs.Rules[k].RuleCategory
+			}
+
 			// default new rule to be resolved if not supplied.
 			if newRule["resolved"] == nil {
 				nr.Resolved = true
@@ -345,6 +427,7 @@ func (rsm ruleSetsModel) GenerateRuleSetFromSuppliedRuleSet(ruleset *RuleSet) *R
 			rs.Rules[k] = &nr
 		}
 	}
+	rs.mutex.Unlock()
 	return rs
 }
 
@@ -354,7 +437,7 @@ func CreateRuleSetFromRuleMap(rules map[string]*model.Rule) *RuleSet {
 	rs := &RuleSet{
 		DocumentationURI: "https://quobix.com/vacuum/rulesets/understanding",
 		Formats:          []string{"oas2", "oas3"},
-		Extends:          map[string]string{SpectralOpenAPI: SpectralOff},
+		Extends:          map[string]string{VacuumOpenAPI: VacuumOff},
 		Description:      fmt.Sprintf("a custom ruleset composed of %d rules", len(rules)),
 		RuleDefinitions:  make(map[string]interface{}),
 		Rules:            rules,
@@ -376,6 +459,7 @@ func GetAllBuiltInRules() map[string]*model.Rule {
 	rules[InfoContact] = GetInfoContactRule()
 	rules[InfoDescription] = GetInfoDescriptionRule()
 	rules[InfoLicense] = GetInfoLicenseRule()
+	rules[InfoLicenseSPDX] = GetInfoLicenseSPDXRule()
 	rules[LicenseUrl] = GetInfoLicenseUrlRule()
 	rules[OpenAPITagsAlphabetical] = GetOpenApiTagsAlphabeticalRule()
 	rules[OpenAPITags] = GetOpenApiTagsRule()
@@ -388,6 +472,7 @@ func GetAllBuiltInRules() map[string]*model.Rule {
 	rules[PathNotIncludeQuery] = GetPathNotIncludeQueryRule()
 	rules[TagDescription] = GetTagDescriptionRequiredRule()
 	rules[NoRefSiblings] = GetNoRefSiblingsRule()
+	rules[Oas3NoRefSiblings] = GetOAS3NoRefSiblingsRule()
 	rules[Oas3UnusedComponent] = GetOAS3UnusedComponentRule()
 	rules[Oas2UnusedDefinition] = GetOAS2UnusedComponentRule()
 	rules[Oas2APIHost] = GetOAS2APIHostRule()
@@ -403,6 +488,7 @@ func GetAllBuiltInRules() map[string]*model.Rule {
 	rules[Oas2OperationSecurityDefined] = GetOAS2SecurityDefinedRule()
 	rules[TypedEnum] = GetTypedEnumRule()
 	rules[DuplicatedEntryInEnum] = GetDuplicatedEntryInEnumRule()
+	rules[NullableEnumRule] = GetNullableEnumRule()
 	rules[NoEvalInMarkdown] = GetNoEvalInMarkdownRule()
 	rules[NoScriptTagsInMarkdown] = GetNoScriptTagsInMarkdownRule()
 	rules[DescriptionDuplication] = GetDescriptionDuplicationRule()
@@ -420,6 +506,14 @@ func GetAllBuiltInRules() map[string]*model.Rule {
 	rules[Oas3ExampleMissingCheck] = GetOAS3ExamplesMissingRule()
 	rules[Oas3ExampleExternalCheck] = GetOAS3ExamplesExternalCheck()
 	rules[OasSchemaCheck] = GetSchemaTypeCheckRule()
+	rules[OasMissingType] = GetMissingTypeRule()
+	rules[PostResponseSuccess] = GetPostSuccessResponseRule()
+	rules[NoRequestBody] = GetNoRequestBodyRule()
+	rules[PathItemReferences] = GetPathItemReferencesRule()
+	rules[DuplicatePathsRule] = GetDuplicatePathsRule()
+	rules[UnnecessaryCombinatorRule] = GetUnnecessaryCombinatorRule()
+	rules[CamelCasePropertiesRule] = GetCamelCasePropertiesRule()
+	rules[MigrateZallyIgnoreRule] = GetMigrateZallyIgnoreRule()
 
 	// dead.
 	//rules[Oas2ValidSchemaExample] = GetOAS2ExamplesRule()
@@ -492,6 +586,7 @@ type RuleSet struct {
 	Rules            map[string]*model.Rule `json:"-" yaml:"-"`
 	Extends          interface{}            `json:"extends,omitempty" yaml:"extends,omitempty"` // can be string or tuple (again... why stoplight?)
 	extendsMeta      map[string]string
+	mutex            sync.Mutex
 }
 
 // GetExtendsValue returns an array of maps defining which ruleset this one extends. The value can be
@@ -531,13 +626,16 @@ func CreateRuleSetUsingJSON(jsonData []byte) (*RuleSet, error) {
 	}
 
 	compiler := jsonschema.NewCompiler()
-	_ = compiler.AddResource("schema.json", strings.NewReader(rulesetSchema))
+
+	// schema needs to be parsed first
+	var parsed map[string]interface{}
+	_ = json.Unmarshal([]byte(RulesetSchema), &parsed)
+	_ = compiler.AddResource("schema.json", parsed)
 	jsch, _ := compiler.Compile("schema.json")
 
 	var data map[string]interface{}
 	_ = json.Unmarshal(jsonData, &data)
 
-	// 4. validate the object against the schema
 	scErrs := jsch.Validate(data)
 
 	if scErrs != nil {
@@ -546,7 +644,7 @@ func CreateRuleSetUsingJSON(jsonData []byte) (*RuleSet, error) {
 		// flatten the validationErrors
 		schFlatErrs := jk.BasicOutput().Errors
 		for q := range schFlatErrs {
-			buf.WriteString(schFlatErrs[q].Error)
+			buf.WriteString(schFlatErrs[q].Error.Kind.LocalizedString(message.NewPrinter(language.Tag{})))
 			if q+1 < len(schFlatErrs) {
 				buf.WriteString(", ")
 			}
@@ -570,6 +668,14 @@ func CreateRuleSetUsingJSON(jsonData []byte) (*RuleSet, error) {
 			if dErr != nil {
 				return nil, dErr
 			}
+			// check if rule has category.
+			if b["category"] != nil {
+				var cat model.RuleCategory
+				dErr = mapstructure.Decode(b["category"], &cat)
+				if dErr == nil {
+					rule.RuleCategory = &cat
+				}
+			}
 			rs.Rules[k] = &rule
 			rule.Resolved = true // default resolved.
 		}
